from base64 import b64encode
from enum import Enum
from typing import List

from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.hazmat.primitives.hashes import SHA256

CUENTA_FIELDNAMES = """
    empresa
    cuenta
    rfcCurp
""".split()


ORDEN_FIELDNAMES = """
    institucionContraparte
    empresa
    fechaOperacion
    folioOrigen
    claveRastreo
    institucionOperante
    monto
    tipoPago
    tipoCuentaOrdenante
    nombreOrdenante
    cuentaOrdenante
    rfcCurpOrdenante
    tipoCuentaBeneficiario
    nombreBeneficiario
    cuentaBeneficiario
    rfcCurpBeneficiario
    emailBeneficiario
    tipoCuentaBeneficiario2
    nombreBeneficiario2
    cuentaBeneficiario2
    rfcCurpBeneficiario2
    conceptoPago
    conceptoPago2
    claveCatUsuario1
    claveCatUsuario2
    clavePago
    referenciaCobranza
    referenciaNumerica
    tipoOperacion
    topologia
    usuario
    medioEntrega
    prioridad
    iva
    """.split()
SIGN_DIGEST = 'RSA-SHA256'


def join_fields(obj: 'Resource', fieldnames: List[str]) -> bytes:  # noqa: F821
    joined_fields = []
    for field in fieldnames:
        value = getattr(obj, field, None)
        if isinstance(value, float):
            value = f'{value:.2f}'
        elif isinstance(value, Enum) and value:
            value = value.value
        elif value is None:
            value = ''
        joined_fields.append(str(value))
    return '||' + '|'.join(joined_fields) + '||'


def compute_signature(pkey: RSAPrivateKey, text: str) -> str:
    signature = pkey.sign(
        text.encode('utf-8'),
        padding.PKCS1v15(),
        SHA256(),
    )
    return b64encode(signature).decode('ascii')



import datetime as dt

import pytest
import requests_mock
from clabe import generate_new_clabes

from stpmex import Client
from stpmex.resources import CuentaFisica, Orden
from stpmex.types import Pais

PKEY = """Bag Attributes
    friendlyName: prueba
    localKeyID: 54 69 6D 65 20 31 33 32 34 35 39 35 30 31 35 33 33 30
Key Attributes: <No Attributes>
-----BEGIN ENCRYPTED PRIVATE KEY-----
MIICxjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQIAPOngEipSGICAggA
MBQGCCqGSIb3DQMHBAi3RX0+96FhJASCAoAGX5N8jxBqlyKk8MTz/Q3a/V4fnCNA
IlPYybMbO00HbMNXw20Kn+WzK73VZtBdEf+8CBcqZMwuC0gqn5pdnOVqP0wz8MU5
AlWu0ZJpLo8npjQyV5Smrk1OvFREQ9skuJRBgYjPPTgdYmVN77ZGeFwJlf+OqOIM
JWZIFZY6z3cXn6CnaAvQ6L+/smRt1Us0gEMe1m7rln0M6m64EbOFsOonzp7/CRTd
Mmlsk93Lg8G/uwGrL3gf1TDep1yM1KKMu6pWZ+6zT26ykwNsdUg0NUCpeWeYWzDZ
KLzQ90U+/XlBPbPg/8gK6tc1dresvPbRcvNu+IJq8HbKuUkjrDeFor5Wezic3CyO
/g//2LJbJGy7Ak4V4W9J46GLD8B3fqyDz0itCBRcmlrtAXiV0azb1isD+j8LdOXN
vo/EPjLJnVdbP2RHiKKdp0Kq2FyRbigP86UujxwxfOxNN/w6m48agmVsj1uB6zBp
hn0D/MLkMtoV7NmGhayRxFXs5sO1G/lWOoR96PgNzOur8xnPzvG7ysPv9qKRO1XS
JEaGZXUUQ/sq2d6nLWMz9YLh7YVaVsRfIcUGPnmFh/bj30Pk52PodF6kN3JYftvn
ZaXgOf6E4NLpHjtYRTzyVZQamenDAlvHQwZE284hDPShuJwxFr6FOSR/GrgqbN4d
cOK898ofM+ZxkNkm5LrU3RAXR3336HU9XMky4UCV9L3CA51IlTMqt/CkddFhsjrw
W4Zo1Aj8G7FaoDm7XhkLGDwVjf0Ua1O4YHRpSgVSkrXeBgW7P4Tc+53nFns3rwxs
uzF/x9tl2+BdiDjPOhSRuoa1ypilODdpOGKNKuf0vu2jAbbzDILBYOfw
-----END ENCRYPTED PRIVATE KEY-----"""


@pytest.fixture
def client():
    empresa = 'TAMIZI'
    pkey_passphrase = '12345678'
    yield Client(empresa, PKEY, pkey_passphrase, demo=True)


@pytest.fixture
def client_mock(request):
    empresa = 'TAMIZI'
    pkey_passphrase = '12345678'

    with requests_mock.mock() as m:
        m.put(requests_mock.ANY, json=request.param)
        yield Client(empresa, PKEY, pkey_passphrase, demo=True)


@pytest.fixture
def orden_dict():
    yield dict(
        institucionContraparte='40072',
        claveRastreo='CR1564969083',
        monto=1.2,
        tipoPago=1,
        nombreOrdenante=None,
        cuentaOrdenante='646180110400000007',
        rfcCurpOrdenante=None,
        nombreBeneficiario='Ricardo Sanchez',
        cuentaBeneficiario='072691004495711499',
        rfcCurpBeneficiario='ND',
        conceptoPago='Prueba',
        referenciaNumerica=5273144,
        topologia='T',
        medioEntrega=3,
        iva=None,
    )


@pytest.fixture
def orden(client, orden_dict):
    yield Orden(**orden_dict)


@pytest.fixture
def cuenta_dict():
    yield dict(
        cuenta=generate_new_clabes(1, '6461801570')[0],
        nombre='Eduardo,Marco',
        apellidoPaterno='Salvador',
        apellidoMaterno='Hernandez-Muñoz',
        rfcCurp='SAHE800416HDFABC01',
        fechaNacimiento=dt.date(1980, 4, 14),
        genero='H',
        entidadFederativa=1,
        actividadEconomica='30',
        calle='mi calle',
        numeroExterior='2',
        numeroInterior='1',
        colonia='mi colonia',
        alcaldiaMunicipio='mi alcaldia',
        cp='12345',
        paisNacimiento=Pais.MX,
        email='asdasd@domain.com',
        idIdentificacion='123123123',
    )


@pytest.fixture
def cuenta(client, cuenta_dict):
    yield CuentaFisica(**cuenta_dict)



import datetime as dt
import random
import time
from dataclasses import field, make_dataclass
from typing import Any, ClassVar, Dict, List, Optional, Union

import clabe
from clabe.types import Clabe
from cuenca_validations.types import (
    PaymentCardNumber,
    StrictPositiveFloat,
    digits,
)
from pydantic import conint, constr, validator
from pydantic.dataclasses import dataclass

from ..auth import ORDEN_FIELDNAMES
from ..exc import NoOrdenesEncontradas
from ..types import (
    BeneficiarioClabe,
    Estado,
    MxPhoneNumber,
    Prioridad,
    TipoCuenta,
    TipoOperacion,
    truncated_str,
)
from ..utils import strftime, strptime
from .base import Resource

STP_BANK_CODE = 90646


@dataclass
class Orden(Resource):
    """
    Based on:
    https://stpmex.zendesk.com/hc/es/articles/360002682851-RegistraOrden-Dispersi%C3%B3n-
    """

    _endpoint: ClassVar[str] = '/ordenPago'
    _firma_fieldnames: ClassVar[List[str]] = ORDEN_FIELDNAMES

    monto: StrictPositiveFloat
    conceptoPago: truncated_str(39)

    cuentaBeneficiario: Union[
        BeneficiarioClabe, PaymentCardNumber, MxPhoneNumber
    ]
    nombreBeneficiario: truncated_str(39)
    institucionContraparte: digits(5, 5)

    cuentaOrdenante: Clabe
    nombreOrdenante: Optional[truncated_str(39)] = None
    institucionOperante: digits(5, 5) = STP_BANK_CODE

    tipoCuentaBeneficiario: Optional[TipoCuenta] = None
    tipoCuentaOrdenante: TipoCuenta = TipoCuenta.clabe.value

    claveRastreo: truncated_str(29) = field(
        default_factory=lambda: f'CR{int(time.time())}'
    )
    referenciaNumerica: conint(gt=0, lt=10 ** 7) = field(
        default_factory=lambda: random.randint(10 ** 6, 10 ** 7)
    )
    rfcCurpBeneficiario: constr(max_length=18) = 'ND'
    rfcCurpOrdenante: Optional[constr(max_length=18)] = None

    prioridad: int = Prioridad.normal.value
    medioEntrega: int = 3
    tipoPago: int = 1
    topologia: str = 'T'
    iva: Optional[float] = None

    id: Optional[int] = None

    def __post_init__(self):
        cb = self.cuentaBeneficiario
        self.tipoCuentaBeneficiario = self.get_tipo_cuenta(cb)

    @classmethod
    def registra(cls, **kwargs) -> 'Orden':
        orden = cls(**kwargs)
        endpoint = orden._endpoint + '/registra'
        resp = orden._client.put(endpoint, orden.to_dict())
        orden.id = resp['id']
        return orden

    @staticmethod
    def get_tipo_cuenta(cuenta: str) -> Optional[TipoCuenta]:
        cuenta_len = len(cuenta)
        if cuenta_len == 18:
            tipo = TipoCuenta.clabe
        elif cuenta_len in {15, 16}:
            tipo = TipoCuenta.card
        elif cuenta_len == 10:
            tipo = TipoCuenta.phone_number
        else:
            tipo = None
        return tipo

    @validator('institucionContraparte')
    def _validate_institucion(cls, v: str) -> str:
        if v not in clabe.BANKS.values():
            raise ValueError(f'{v} no se corresponde a un banco')
        return v

    @classmethod
    def consulta_recibidas(
        cls, fecha_operacion: Optional[dt.date] = None
    ) -> List['OrdenConsultada']:  # noqa: F821
        """
        Consultar
        """
        return cls._consulta_fecha(TipoOperacion.recibida, fecha_operacion)

    @classmethod
    def consulta_enviadas(
        cls, fecha_operacion: Optional[dt.date] = None
    ) -> List['OrdenConsultada']:  # noqa: F821
        return cls._consulta_fecha(TipoOperacion.enviada, fecha_operacion)

    @classmethod
    def consulta_clave_rastreo(
        cls,
        claveRastreo: str,
        institucionOperante: Union[int, str],
        fechaOperacion: Optional[dt.date] = None,
    ) -> 'OrdenConsultada':  # noqa: F821
        """
        Consultar ordenes por clave rastreo. Exclude the fechaOperacion if
        looking up transactions from the same day or when the fechaOperacion is
        in the future, in the event of this function being called during
        non-banking hours (9am – 6pm) / days.

        Based on:
        https://stpmex.zendesk.com/hc/es/articles/360039782292-Consulta-Orden-Enviada-Por-Rastreo
        """
        institucionOperante = int(institucionOperante)
        if institucionOperante == STP_BANK_CODE:  # enviada
            consulta_method = cls._consulta_clave_rastreo_enviada

        else:  # recibida
            consulta_method = cls._consulta_clave_rastreo_recibida
        return consulta_method(
            claveRastreo, institucionOperante, fechaOperacion
        )

    @classmethod
    def _consulta_fecha(
        cls, tipo: TipoOperacion, fechaOperacion: Optional[dt.date] = None
    ) -> List['OrdenConsultada']:  # noqa: F821
        """
        Exclude the fechaOperacion if looking up transactions from the same
        day or when the fechaOperacion is in the future, in the event of this
        function being called during non-banking hours (9am – 6pm) / days.
        """
        endpoint = cls._endpoint + '/consOrdenesFech'
        consulta = dict(empresa=cls.empresa, estado=tipo)
        if fechaOperacion:
            consulta['fechaOperacion'] = strftime(fechaOperacion)
        consulta['firma'] = cls._firma_consulta(consulta)
        try:
            resp = cls._client.post(endpoint, consulta)
        except NoOrdenesEncontradas:
            ordenes = []
        else:
            ordenes = [
                cls._sanitize_consulta(orden) for orden in resp['lst'] if orden
            ]
        return ordenes

    @classmethod
    def _consulta_clave_rastreo_enviada(
        cls,
        claveRastreo: str,
        institucionOperante: int,
        fechaOperacion: Optional[dt.date] = None,
    ) -> 'OrdenConsultada':  # noqa: F821
        endpoint = cls._endpoint + '/consOrdEnvRastreo'
        consulta = dict(
            empresa=cls.empresa,
            claveRastreo=claveRastreo,
            institucionOperante=institucionOperante,
        )
        if fechaOperacion:
            consulta['fechaOperacion'] = strftime(fechaOperacion)
        consulta['firma'] = cls._firma_consulta(consulta)
        resp = cls._client.post(endpoint, consulta)['ordenPago']
        return cls._sanitize_consulta(resp)

    @classmethod
    def _consulta_clave_rastreo_recibida(
        cls,
        claveRastreo: str,
        institucionOperante: int,
        fechaOperacion: Optional[dt.date] = None,
    ) -> 'OrdenConsultada':  # noqa: F821
        recibidas = cls.consulta_recibidas(fechaOperacion)
        orden = None
        for o in recibidas:
            if o.claveRastreo == claveRastreo and institucionOperante in (
                o.institucionOperante,
                o.institucionContraparte,
            ):
                orden = o
                break
        if not orden:
            raise NoOrdenesEncontradas
        return orden

    @staticmethod
    def _sanitize_consulta(
        orden: Dict[str, Any]
    ) -> 'OrdenConsultada':  # noqa: F821
        sanitized = {}
        for k, v in orden.items():
            if k.startswith('ts'):
                v /= 10 ** 3  # convertir de milisegundos a segundos
                if v > 10 ** 9:
                    v = dt.datetime.fromtimestamp(v)
            elif k == 'fechaOperacion':
                v = strptime(v)
            elif k == 'estado':
                v = Estado(v)
            elif isinstance(v, str):
                v = v.rstrip()
            sanitized[k] = v
        return make_dataclass('OrdenConsultada', sanitized.keys())(**sanitized)



import datetime as dt

from stpmex.utils import strftime, strptime


def test_strftime():
    assert strftime(dt.date(2020, 4, 20)) == '20200420'


def test_strptime():
    assert strptime('20200420') == dt.date(2020, 4, 20)



import datetime as dt
from typing import Union

DATE_FORMAT = '%Y%m%d'


def strftime(date: dt.date):
    return date.strftime(DATE_FORMAT)


def strptime(date: Union[int, str]):
    return dt.datetime.strptime(str(date), DATE_FORMAT).date()



from pydantic import PydanticValueError


class StpmexException(Exception):
    def __init__(self, **kwargs):
        for attr, value in kwargs.items():
            setattr(self, attr, value)

    def __repr__(self):
        return (
            self.__class__.__name__
            + '('
            + ', '.join(
                [
                    f'{attr}={repr(value)}'
                    for attr, value in self.__dict__.items()
                ]
            )
            + ')'
        )

    def __str__(self):
        return repr(self)


class InvalidPassphrase(StpmexException):
    """El passphrase es incorrecto"""


class InvalidAccountType(StpmexException):
    """Tipo de cuenta inválida"""


class SignatureValidationError(StpmexException):
    """Error validando la firma"""


class InvalidRfcOrCurp(StpmexException):
    """RFC o CURP inválido"""


class ClaveRastreoAlreadyInUse(StpmexException):
    """La clave de rastreo es repetida"""


class PldRejected(StpmexException):
    """'Orden sin cuenta ordenante. Se rechaza por PLD"""


class NoServiceResponse(StpmexException):
    """No se recibió respuesta del servicio"""


class NoOrdenesEncontradas(StpmexException):
    """No se encontraron ordenes"""


class InvalidTrackingKey(StpmexException):
    """Clave de rastreeo inválida"""


class BankCodeClabeMismatch(StpmexException):
    """La cuenta clabe no coincide para la institución operante"""


class SameAccount(StpmexException):
    """Transferencia a la misma cuenta"""


class DuplicatedAccount(StpmexException):
    """Cuenta duplicada"""


class InvalidField(StpmexException):
    """Campo inválido"""


class MandatoryField(StpmexException):
    """El campo X es obligatorio"""


class InvalidInstitution(StpmexException):
    """La Institucion no es valida"""


class AccountDoesNotExist(StpmexException):
    """La cuenta no existe"""


class InvalidAmount(StpmexException):
    """El monto es inválido para una de las instituciones"""


class BlockedInstitutionError(PydanticValueError):
    """Institución bloqueada"""

    code = 'clabe.bank_code'
    msg_template = '{bank_name} has been blocked by STP.'



from stpmex.auth import (
    CUENTA_FIELDNAMES,
    ORDEN_FIELDNAMES,
    compute_signature,
    join_fields,
)


def test_join_fields_for_orden(orden):
    joined = (
        '||40072|TAMIZI|||CR1564969083|90646|1.20|1|40||646180110400000007|'
        '|40|Ricardo Sanchez|072691004495711499|ND||||||Prueba||||||5273144|'
        '|T||3|0|||'
    )
    assert join_fields(orden, ORDEN_FIELDNAMES) == joined


def test_join_fields_for_cuenta(cuenta):
    cuenta.cuenta = '646180157099999993'
    joined = '||TAMIZI|646180157099999993|SAHE800416HDFABC01||'
    assert join_fields(cuenta, CUENTA_FIELDNAMES) == joined


def test_compute_signature(client, orden):
    firma = (
        'KDNKDVVuyNt9oTXPAlofGXGH5L5IH9PAzOsx0JZFtmGlU+10QRf2RHSg0OVCnYYpu5sC3'
        'DJ6vlXuYM40+uNw0tMc0y8Dv26uO8Vv2GhOhMqaGk72LwgwgmqVg17xzjgGbJHzAzMav3'
        'fx4/3No+mSnf7vxpe4ePf6yK1yU5U28L4='
    )
    sig = compute_signature(client.pkey, join_fields(orden, ORDEN_FIELDNAMES))
    assert sig == firma



import re
import unicodedata
from enum import Enum
from typing import TYPE_CHECKING, ClassVar, Type

from clabe import Clabe
from cuenca_validations.validators import validate_digits
from pydantic import ConstrainedStr, StrictStr
from pydantic.validators import (
    constr_length_validator,
    constr_strip_whitespace,
    str_validator,
)

from stpmex.exc import BlockedInstitutionError

if TYPE_CHECKING:
    from pydantic.typing import CallableGenerator

# STP does not allow to make tranfers to this banks codes.
BLOCKED_INSTITUTIONS = {'90659', '90642'}


def unicode_to_ascii(unicode: str) -> str:
    v = unicodedata.normalize('NFKD', unicode).encode('ascii', 'ignore')
    return v.decode('ascii')


class AsciiStr(ConstrainedStr):
    @classmethod
    def __get_validators__(cls) -> 'CallableGenerator':
        yield unicode_to_ascii
        yield from super().__get_validators__()
        yield lambda value: value.strip()


class StpStr(AsciiStr):
    """
    based on:
    https://stpmex.zendesk.com/hc/es/articles/360038242071-Registro-de-Cuentas-de-Personas-f%C3%ADsicas
    """

    @classmethod
    def validate(cls, value: str) -> str:
        value = super().validate(value)
        value = re.sub(r'[-,.]', ' ', value)
        value = value.upper()
        return value


class BeneficiarioClabe(Clabe):
    @classmethod
    def __get_validators__(cls) -> 'CallableGenerator':
        yield from Clabe.__get_validators__()
        yield cls.validate_blocked_institution

    @classmethod
    def validate_blocked_institution(cls, clabe: Clabe) -> Clabe:
        if clabe.bank_code_banxico in BLOCKED_INSTITUTIONS:
            raise BlockedInstitutionError(bank_name=clabe.bank_name)
        return clabe


def truncated_str(length: int) -> Type[str]:
    namespace = dict(
        strip_whitespace=True, min_length=1, curtail_length=length
    )
    return type('TruncatedStrValue', (AsciiStr,), namespace)


def truncated_stp_str(length: int) -> Type[str]:
    namespace = dict(
        strip_whitespace=True, min_length=1, curtail_length=length
    )
    return type('TruncatedStpStrValue', (StpStr,), namespace)


class Estado(str, Enum):
    """
    Based on: https://stpmex.zendesk.com/hc/es/articles/360040200791
    """

    capturada = 'C'
    pendiente_liberar = 'PL'
    liberada = 'L'
    pendiente_autorizar = 'PA'
    autorizada = 'A'
    enviada = 'E'
    liquidada = 'LQ'
    cancelada = 'CN'
    traspaso_liberado = 'TL'
    traspaso_capturado = 'TC'
    traspaso_autorizado = 'TA'
    traspaso_liquidado = 'TLQ'
    traspaso_cancelado = 'TCL'
    recibida = 'R'
    por_devolver = 'XD'
    devuelta = 'D'
    por_enviar_confirmacion = 'CXO'
    confirmacion_enviada = 'CCE'
    confirmada = 'CCO'
    confirmacion_rechazada = 'CCR'
    por_cancelar = 'XC'
    cancelada_local = 'CL'
    cancelada_rechazada = 'CR'
    rechazada_local = 'RL'
    cancelada_adapter = 'CA'
    rechazada_adapter = 'RA'
    enviada_adapter = 'EA'
    rechazada_banxico = 'RB'
    eliminada = 'EL'
    por_retornar = 'XR'
    retornada = 'RE'
    exportacion_poa = 'EP'
    exportacion_cep = 'EC'


class Prioridad(int, Enum):
    normal = 0
    alta = 1


class TipoCuenta(int, Enum):
    card = 3
    phone_number = 10
    clabe = 40


class Genero(str, Enum):
    mujer = 'M'
    hombre = 'H'


class Curp(StrictStr):
    min_length = 18
    max_length = 18
    regex = re.compile(r'^[A-Z]{4}[0-9]{6}[A-Z]{6}[A-Z|0-9][0-9]$')


class Rfc(StrictStr):
    min_length = 12
    max_length = 13


class EntidadFederativa(int, Enum):
    # NE = Nacido en el Extranjero. Aún STP no soporte
    AS = 1  # Aguascalientes
    BC = 2  # Baja California
    BS = 3  # Baja California Sur
    CC = 4  # Campeche
    CS = 5  # Chiapas
    CH = 6  # Chihuahua
    CL = 7  # Coahuila
    CM = 8  # Colima
    DF = 9  # CDMX
    DG = 10  # Durango
    MC = 11  # Estado de México
    GT = 12  # Guanajuato
    GR = 13  # Guerrero
    HG = 14  # Hidalgo
    JC = 15  # Jalisco
    MN = 16  # Michoacan
    MS = 17  # Morelos
    NT = 18  # Nayarit
    NL = 19  # Nuevo León
    OC = 20  # Oaxaca
    PL = 21  # Puebla
    QT = 22  # Querétaro
    QR = 23  # Quintana Roo
    SP = 24  # San Luis Potosí
    SL = 25  # Sinaloa
    SR = 26  # Sonora
    TC = 27  # Tabasco
    TS = 28  # Tamualipas
    TL = 29  # Tlaxcala
    VZ = 30  # Veracruz
    YN = 31  # Yucatán
    ZS = 32  # Zacatecas


class Pais(int, Enum):
    """
    Based on https://stpmex.zendesk.com/hc/es/articles/360037876272
    """

    SE_DESCONOCE = 0
    AF = 1  # Republica Islamica de Afganistan
    AL = 2  # Republica de Albania
    DE = 3  # Republica Federal de Alemania
    HV = 4  # Alto Volta
    AD = 5  # Principado de Andorra
    AO = 6  # Republica de Angola
    AI = 7  # Anguila
    AG = 8  # Antigua y Barbuda
    AN = 9  # Antillas Neerlandesas
    SA = 10  # Reino de Arabia Saudita
    SJ = 11  # Svalbard y Jan Mayen
    DZ = 12  # Republica Democratica Popular de Argelia
    AR = 13  # Republica Argentina
    AM = 14  # Republica de Armenia
    AW = 15  # Aruba
    AC = 16  # Islas de Ascencion
    AU = 17  # Commonwealth de Australia
    AT = 18  # Republica de Austria
    AZ = 19  # Republica de Azerbaiyan
    BS = 20  # Commonwealth de Las Bahamas
    BH = 21  # Reino de Bahrein
    BD = 22  # Republica Popular de Bangladesh
    BB = 23  # Barbados
    BY = 24  # Republica de Belarus
    BE = 25  # Reino de Belgica
    BZ = 26  # Belice
    BM = 27  # Bermudas
    MM = 28  # Birmania
    BO = 29  # Republica de Bolivia
    BA = 30  # Bosnia Herzegovina
    BW = 31  # Republica de Botswana
    BR = 32  # Republica Federal de Brasil
    BN = 33  # Brunei Malasia
    BG = 34  # Republica de Bulgaria
    BI = 35  # Republica de Burundi
    BT = 36  # Reino de Butan
    CM = 37  # Republica de Camerun
    CA = 39  # Dominio de Canada
    CO = 40  # Republica de Colombia
    KR = 43  # Republica de Corea
    CI = 44  # Republica de Costa de Marfil
    CR = 45  # Republica de Costa Rica
    HR = 46  # Republica de Croacia
    CU = 47  # Republica de Cuba
    CW = 48  # Curazao
    TD = 49  # Republica de Chad
    CZ = 50  # Republica Checa
    CL = 51  # Republica de Chile
    CN = 52  # Republica Popular China
    DK = 54  # Reino de Dinamarca
    EC = 56  # Republica del Ecuador
    EG = 57  # Republica Arabe de Egipto
    SV = 58  # Republica de El Salvador
    AE = 59  # Emiratos Arabes Unidos
    ES = 60  # Reino de Espana
    KW = 61  # Estado de Kuwait
    QA = 62  # Estado de Qatar
    US = 63  # Estados Unidos de Norteamerica
    EE = 64  # Republica de Estonia
    ET = 65  # Republica Democratica Federal de Etiopia
    PH = 66  # Republica de Las Filipinas
    FI = 67  # Republica de Finlandia
    FR = 68  # Republica de Francia
    GA = 69  # Republica de Gabon
    GM = 70  # Republica de La Gambia
    GE = 71  # Georgia
    GH = 72  # Republica de Ghana
    GI = 73  # Gibraltar
    GD = 74  # Granada
    GR = 75  # Grecia
    GL = 76  # Groenlandia
    GU = 77  # Guam
    GT = 78  # Guatemala
    GF = 79  # Guayana Francesa
    GN = 80  # Guinea
    GQ = 81  # Guinea Ecuatorial
    GY = 83  # Guyana
    HT = 86  # Republica de Haiti
    NL = 87  # Holanda
    HN = 88  # Republica de Honduras
    HK = 89  # Hong Kong
    HU = 90  # Hungria
    IN = 91  # Republica de India
    ID = 92  # Republica de Indonesia
    GB = 93  # Reino Unido
    IQ = 94  # Republica de Irak
    IR = 95  # Republica Islamica de Iran
    IE = 96  # Republica de Irlanda
    KY = 97  # Islas Caiman
    NF = 98  # Isla de Norfolk
    PM = 100  # Isla de San Pedro y Miquelin
    IM = 101  # Isla de Man
    IS = 103  # Islandia
    IC = 105  # Islas Canarias
    CK = 106  # Islas Cook
    CC = 107  # Islas de Cocos O Kelling
    GG = 108  # Guernesey
    FK = 109  # Islas Malvinas
    MH = 110  # Republica de Las Islas Marshall
    SB = 112  # Islas Salomon
    TC = 113  # Islas Turcas y Caicos
    VG = 114  # Islas Virgenes Britanicas
    IL = 116  # Estado de Israel
    IT = 117  # Republica de Italia
    JM = 118  # Jamaica
    JP = 119  # Japon
    JO = 120  # Reino Hashemita de Jordania
    KZ = 121  # Republica de Kazajstan
    KE = 122  # Republica de Kenya
    KG = 123  # Republica de Kirguistan
    KI = 124  # Republica de Kiribati
    LA = 127  # Republica Democratica Popular de Laos
    LS = 129  # Reino de Lesotho
    LB = 130  # Republica del Libano
    LR = 131  # Republica de Liberia
    LI = 133  # Principado de Liechenstein
    LT = 134  # Republica de Lituania
    LU = 135  # Gran Ducado de Luxemburgo
    MO = 136  # Region Especialadminiostrativademacaodelarepublicapopularchina
    YU = 137  # Antigua Republica Yugoslava de Macedonia
    MG = 138  # Republica de Madagascar
    MY = 140  # Malasia
    MW = 141  # Republica de Malawi
    ML = 142  # Republica de Mali
    MT = 143  # Republica de Malta
    MA = 144  # Reino de Marruecos
    MU = 145  # Republica de Mauricio
    MR = 146  # Republica Islamica de Mauritania
    MC = 148  # Principado de Monaco
    MN = 149  # Mongolia
    MS = 150  # Montserrat
    MZ = 151  # Republica de Mozambique
    NA = 152  # Republica de Namibia
    NR = 153  # Republica de Nauru
    NP = 154  # Estado de Nepal
    NI = 156  # Republica de Nicaragua
    NE = 268  # Niger
    NG = 158  # Republica Federal de Nigeria
    NU = 159  # Niue
    NO = 161  # Reino de Noruega
    NZ = 162  # Nueva Zelanda
    OM = 163  # Sultanato de Oman
    PW = 164  # Republica de Palaos
    PA = 165  # Republica Panama
    PK = 166  # Republica Islamica de Paquistan
    PY = 167  # Republica de Paraguay
    PE = 168  # Republica del Peru
    PN = 169  # Islas Pitcairn
    PF = 170  # Polinesia Francesa
    PL = 171  # Republica de Polonia
    PT = 172  # Republica de Portugal
    PR = 173  # Estado Libre Asociado de Puerto Rico
    TO = 175  # Reino de Tonga
    CV = 177  # Republica de Cabo Verde
    CY = 178  # Republica de Chipre
    MV = 181  # Republica de Las Maldivas
    SC = 182  # Republica de Seychelles
    TN = 183  # Republica de Tunez
    VU = 184  # Republica de Vanuatu
    YE = 185  # Republica del Yemen
    DO = 186  # Republica Dominicana
    MX = 187  # Mexico
    UY = 188  # Republica Oriental de Uruguay
    RW = 190  # Republica de Ruanda
    RO = 191  # Rumania
    RU = 192  # Federacion Rusa, Rusia
    AS = 194  # Samoa Americana
    VC = 197  # San Vicente y Las Granadinas
    SN = 199  # Republica de Senegal
    SL = 201  # Republica de Sierra Leona
    SK = 203  # Slovakia
    SI = 204  # Slovenia
    SO = 205  # Somalia
    LK = 206  # Republica Democratica Socialista de Sri Lanka
    ZA = 207  # Republica de Sudafrica
    SD = 208  # Republica del Sudan
    SE = 209  # Reino de Suecia
    CH = 210  # Confederacion Helvetica, Suiza
    SR = 211  # Republica de Surinam
    TH = 212  # Reino de Tailandia
    TW = 213  # Republica de China, Taiwan
    TZ = 214  # Republica Unida de Tanzania
    TJ = 215  # Republica de Tajikistan
    TG = 216  # Republica de Togo
    TK = 217  # Tokelau
    TT = 219  # Republica de Trinidad y Tobago
    SH = 220  # Tristan de Cunha
    TM = 222  # Turkmenistan
    TR = 223  # Republica de Turquia
    TV = 224  # Tuvalu
    UA = 225  # Ucrania
    UG = 226  # Uganda
    UZ = 227  # Republica de Uzbekistan
    VE = 228  # Republica Bolivariana de Venezuela
    VN = 229  # Republica Socialista de Vietnam
    ZR = 231  # Republica de Zaire
    ZM = 232  # Republica de Zambia
    KN = 235  # San Cristobal y Nieves
    SG = 236  # Republica de Singapure
    WS = 237  # Samoa
    BF = 238  # Burkina Faso
    CX = 240  # Isla de Christmas
    KM = 241  # Comoras
    ER = 242  # Eritrea
    FO = 243  # Islas Faroe
    FJ = 244  # Fiyi
    MQ = 245  # Martinica
    FM = 246  # Micronesia
    MD = 247  # Moldova
    ME = 248  # Montenegro
    NC = 249  # Nueva Caledonia
    PS = 250  # Palestina
    PG = 251  # Papua Nueva Guinea
    SZ = 252  # Suazilandia
    ST = 253  # Santo Tome y Principe
    RS = 254  # Serbia
    ZW = 255  # Zimbabue
    AQ = 256  # Antartida
    CD = 259  # Republica Democratica del Congo
    AX = 257  # Islas Åland
    BV = 258  # Isla Bouvet
    CP = 260  # Clipperton
    GP = 261  # Guadalupe
    GS = 262  # Georgia del Sur E Islas Sandwich del Sur
    HM = 263  # Islas Heard y Mcdonald
    IO = 264  # Territorio Britanico del Oceano Indico
    JE = 265  # Jersey
    LV = 266  # Letonia
    MP = 267  # Islas Marianas del Norte
    RE = 270  # Reunion
    TF = 271  # Territorios Australes Franceses
    UM = 272  # Islas Menores Alejadas de Los Estados Unidos
    VA = 273  # Santa Sede Estado de La Ciudad del Vaticano
    WF = 274  # Wallis y Futuna
    YT = 275  # Mayotte


class TipoOperacion(str, Enum):
    enviada = 'E'
    recibida = 'R'


class MxPhoneNumber(str):
    strip_whitespace: ClassVar[bool] = True
    min_length: ClassVar[int] = 10
    max_length: ClassVar[int] = 10

    @classmethod
    def __get_validators__(cls) -> 'CallableGenerator':
        yield str_validator
        yield constr_strip_whitespace
        yield constr_length_validator
        yield validate_digits



import datetime as dt
from typing import Any, ClassVar, Dict, List, Optional, Union

from clabe.types import Clabe
from cuenca_validations.types import digits
from pydantic import conint, constr
from pydantic.dataclasses import dataclass

from ..auth import CUENTA_FIELDNAMES
from ..types import (
    Curp,
    EntidadFederativa,
    Genero,
    MxPhoneNumber,
    Pais,
    Rfc,
    truncated_stp_str,
)
from .base import Resource

MAX_LOTE = 100


@dataclass
class Cuenta(Resource):
    _base_endpoint: ClassVar[str] = '/cuentaModule'
    _lote_endpoint: ClassVar[str]
    _firma_fieldnames: ClassVar[List[str]] = CUENTA_FIELDNAMES

    cuenta: Clabe
    rfcCurp: Union[Curp, Rfc]

    @classmethod
    def alta(cls, **kwargs) -> 'Cuenta':
        cuenta = cls(**kwargs)
        cuenta._alta()
        return cuenta

    def _alta(self) -> None:
        self._client.put(self._endpoint, self.to_dict())

    @classmethod
    def alta_lote(cls, lote: List['Cuenta']) -> Dict[str, Dict[str, Any]]:
        if len(lote) > MAX_LOTE:
            return {
                **cls.alta_lote(lote[:MAX_LOTE]),
                **cls.alta_lote(lote[MAX_LOTE:]),
            }
        cuentas = dict(cuentasFisicas=[cuenta.to_dict() for cuenta in lote])
        return dict(
            zip(
                [cuenta.cuenta for cuenta in lote],
                cls._client.put(cls._lote_endpoint, cuentas),
            )
        )

    def baja(self, endpoint: Optional[str] = None) -> Dict[str, Any]:
        endpoint = endpoint or self._endpoint
        data = dict(
            cuenta=self.cuenta,
            empresa=self.empresa,
            rfcCurp=self.rfcCurp,
            firma=self.firma,
        )
        return self._client.delete(endpoint, data)


@dataclass
class CuentaFisica(Cuenta):
    """
    Based on:
    https://stpmex.zendesk.com/hc/es/articles/360038242071-Registro-de-Cuentas-de-Personas-f%C3%ADsicas
    """

    _endpoint: ClassVar[str] = Cuenta._base_endpoint + '/fisica'
    _lote_endpoint: ClassVar[str] = Cuenta._base_endpoint + '/fisicas'

    nombre: truncated_stp_str(50)
    apellidoPaterno: truncated_stp_str(50)
    paisNacimiento: Pais
    fechaNacimiento: dt.date

    apellidoMaterno: Optional[truncated_stp_str(50)] = None
    genero: Optional[Genero] = None
    # Esperando a que STP agregue Nacido en el Extranjero
    entidadFederativa: Optional[EntidadFederativa] = None
    actividadEconomica: Optional[conint(ge=28, le=74)] = None
    calle: Optional[truncated_stp_str(60)] = None
    numeroExterior: Optional[truncated_stp_str(10)] = None
    numeroInterior: Optional[truncated_stp_str(5)] = None
    colonia: Optional[truncated_stp_str(50)] = None
    alcaldiaMunicipio: Optional[truncated_stp_str(50)] = None
    cp: Optional[digits(5, 5)] = None
    email: Optional[constr(max_length=150)] = None
    idIdentificacion: Optional[digits(max_length=20)] = None
    telefono: Optional[MxPhoneNumber] = None



from typing import ClassVar, List
from xml.etree import ElementTree

from pydantic import PositiveFloat, PositiveInt
from pydantic.dataclasses import dataclass

from ..auth import compute_signature
from ..types import TipoOperacion
from .base import Resource


@dataclass
class Saldo(Resource):
    _endpoint: ClassVar[str] = '/ordenPago/consSaldoEnvRec'

    montoTotal: PositiveFloat
    tipoOperacion: TipoOperacion
    totalOperaciones: PositiveInt

    @classmethod
    def consulta_saldo_env_rec(cls) -> List['Saldo']:
        data = dict(empresa=cls.empresa, firma=cls._firma_consulta({}))
        resp = cls._client.post(cls._endpoint, data)
        saldos = []
        for saldo in resp['saldos']:
            del saldo['empresa']
            saldos.append(cls(**saldo))
        return saldos

    @classmethod
    def consulta(cls, cuenta: str) -> float:
        """
        cuenta es la CLABE de la empresa

        Based on:
        https://stpmex.zendesk.com/hc/es/articles/360002812571-consultaSaldoCuenta
        """
        client = cls._client
        firma = compute_signature(client.pkey, cuenta)
        data = f'''
<soapenv:Envelope
        xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/"
        xmlns:h2h="http://h2h.integration.spei.enlacefi.lgec.com/">
    <soapenv:Body>
        <h2h:consultaSaldoCuenta>
            <cuenta>{cuenta}</cuenta>
            <firma>{firma}</firma>
        </h2h:consultaSaldoCuenta>
    </soapenv:Body>
</soapenv:Envelope>
'''
        resp = client.session.post(client.soap_url, data)
        if not resp.ok:
            resp.raise_for_status()
        root = ElementTree.fromstring(resp.text)
        saldo = root.findtext('.//saldo')
        return float(saldo)



import clabe
import pytest
from pydantic import ValidationError

from stpmex.resources import Orden

ORDEN_KWARGS = dict(
    institucionContraparte='40072',
    claveRastreo='CR1564969083',
    monto=1.2,
    tipoPago=1,
    cuentaOrdenante='646180110400000007',
    nombreBeneficiario='Ricardo Sanchez',
    cuentaBeneficiario='072691004495711499',
    conceptoPago='Prueba',
    referenciaNumerica=5273144,
)


def create_orden(**kwargs) -> Orden:
    return Orden(**{**ORDEN_KWARGS, **kwargs})


def test_empty_concepto():
    with pytest.raises(ValidationError) as exc_info:
        create_orden(conceptoPago=' ')
    errors = exc_info.value.errors()
    assert len(errors) == 1
    error = errors[0]
    assert error['loc'] == ('conceptoPago',)
    assert error['type'] == 'value_error.any_str.min_length'


def test_strip_spaces():
    orden = create_orden(conceptoPago=' hello ')
    assert orden.conceptoPago == 'hello'


def test_nonfloat_monto():
    with pytest.raises(ValueError):
        create_orden(monto=5)


def test_truncate_nombre_beneficiario():
    nombre = 'x' * 50
    orden = create_orden(nombreBeneficiario=nombre)
    assert len(orden.nombreBeneficiario) == 39


def test_leading_and_trailing_spaces_nombre_beneficiario():
    nombre = f' {"x" * 38} {"x" * 30} '
    orden = create_orden(nombreOrdenante=nombre)
    assert orden.nombreOrdenante == "x" * 38


def test_invalid_clabe():
    invalid_clabe = '072691004495711490'
    assert not clabe.validate_clabe(invalid_clabe)
    with pytest.raises(ValidationError) as exc_info:
        create_orden(cuentaBeneficiario=invalid_clabe)
    errors = exc_info.value.errors()
    assert len(errors) == 3
    error = errors[0]
    assert error['loc'][0] == 'cuentaBeneficiario'
    assert error['type'] == 'value_error.clabe.control_digit'


def test_wrong_length_cuentaBeneficiario():
    with pytest.raises(ValidationError) as exc_info:
        create_orden(cuentaBeneficiario='1' * 14)
    errors = exc_info.value.errors()
    assert len(errors) == 3
    assert errors[0]['type'] == 'value_error.any_str.min_length'
    assert errors[1]['type'] == 'value_error.any_str.min_length'
    assert errors[2]['type'] == 'value_error.any_str.max_length'
    for error in errors:
        assert error['loc'][0] == 'cuentaBeneficiario'


def test_invalid_bank():
    with pytest.raises(ValidationError) as exc_info:
        create_orden(institucionContraparte='11111')
    errors = exc_info.value.errors()
    assert len(errors) == 1
    error = errors[0]
    assert error['loc'] == ('institucionContraparte',)
    assert error['type'] == 'value_error'


def test_replace_unicode():
    orden = create_orden(
        nombreBeneficiario='Ricardo Sánchez', conceptoPago='está bien, güey 💙'
    )
    assert orden.nombreBeneficiario == 'Ricardo Sanchez'
    assert orden.conceptoPago == 'esta bien, guey'


def test_raises_validation_error_concepto_only_emojies():
    with pytest.raises(ValidationError) as exc_info:
        create_orden(nombreBeneficiario='Ricardo Sánchez', conceptoPago='😁✌️')

    errors = exc_info.value.errors()
    assert len(errors) == 1
    assert errors[0]['loc'] == ('conceptoPago',)
    assert errors[0]['type'] == 'value_error.any_str.min_length'


def test_defaults():
    orden_kwargs = ORDEN_KWARGS.copy()
    orden_kwargs.pop('claveRastreo')
    orden_kwargs.pop('referenciaNumerica')
    orden = Orden(**orden_kwargs)
    assert orden.claveRastreo
    assert orden.referenciaNumerica


def test_zero_referencia_numerica():
    with pytest.raises(ValidationError) as exc_info:
        create_orden(referenciaNumerica='00')
    errors = exc_info.value.errors()
    assert len(errors) == 1
    error = errors[0]
    assert error['loc'] == ('referenciaNumerica',)
    assert error['type'] == 'value_error.number.not_gt'


def test_referencia_numerica_too_high():
    with pytest.raises(ValidationError) as exc_info:
        create_orden(referenciaNumerica=10 ** 7)
    errors = exc_info.value.errors()
    assert len(errors) == 1
    error = errors[0]
    assert error['loc'] == ('referenciaNumerica',)
    assert error['type'] == 'value_error.number.not_lt'


def test_valid_card_number_cuenta_beneficiario():
    orden = create_orden(cuentaBeneficiario='5339220423090005')
    assert orden.claveRastreo
    assert orden.referenciaNumerica


def test_invalid_card_number_cuenta_beneficiario():
    with pytest.raises(ValidationError) as exc_info:
        create_orden(cuentaBeneficiario='5339220423090006')
    errors = exc_info.value.errors()
    assert len(errors) == 3

    error = errors[0]
    assert error['loc'] == ('cuentaBeneficiario',)
    assert error['type'] == 'value_error.any_str.min_length'

    error = errors[1]
    assert error['loc'] == ('cuentaBeneficiario',)
    assert error['type'] == 'value_error.payment_card_number.luhn_check'

    error = errors[2]
    assert error['loc'] == ('cuentaBeneficiario',)
    assert error['type'] == 'value_error.any_str.max_length'



import pytest
from requests import HTTPError

from stpmex.client import Client
from stpmex.exc import (
    AccountDoesNotExist,
    BankCodeClabeMismatch,
    ClaveRastreoAlreadyInUse,
    DuplicatedAccount,
    InvalidAccountType,
    InvalidAmount,
    InvalidField,
    InvalidInstitution,
    InvalidPassphrase,
    InvalidRfcOrCurp,
    InvalidTrackingKey,
    MandatoryField,
    NoServiceResponse,
    PldRejected,
    SameAccount,
    SignatureValidationError,
    StpmexException,
)

PKEY = """Bag Attributes
    friendlyName: prueba
    localKeyID: 54 69 6D 65 20 31 33 32 34 35 39 35 30 31 35 33 33 30
Key Attributes: <No Attributes>
-----BEGIN ENCRYPTED PRIVATE KEY-----
MIICxjBABgkqhkiG9w0BBQ0wMzAbBgkqhkiG9w0BBQwwDgQIAPOngEipSGICAggA
MBQGCCqGSIb3DQMHBAi3RX0+96FhJASCAoAGX5N8jxBqlyKk8MTz/Q3a/V4fnCNA
IlPYybMbO00HbMNXw20Kn+WzK73VZtBdEf+8CBcqZMwuC0gqn5pdnOVqP0wz8MU5
AlWu0ZJpLo8npjQyV5Smrk1OvFREQ9skuJRBgYjPPTgdYmVN77ZGeFwJlf+OqOIM
JWZIFZY6z3cXn6CnaAvQ6L+/smRt1Us0gEMe1m7rln0M6m64EbOFsOonzp7/CRTd
Mmlsk93Lg8G/uwGrL3gf1TDep1yM1KKMu6pWZ+6zT26ykwNsdUg0NUCpeWeYWzDZ
KLzQ90U+/XlBPbPg/8gK6tc1dresvPbRcvNu+IJq8HbKuUkjrDeFor5Wezic3CyO
/g//2LJbJGy7Ak4V4W9J46GLD8B3fqyDz0itCBRcmlrtAXiV0azb1isD+j8LdOXN
vo/EPjLJnVdbP2RHiKKdp0Kq2FyRbigP86UujxwxfOxNN/w6m48agmVsj1uB6zBp
hn0D/MLkMtoV7NmGhayRxFXs5sO1G/lWOoR96PgNzOur8xnPzvG7ysPv9qKRO1XS
JEaGZXUUQ/sq2d6nLWMz9YLh7YVaVsRfIcUGPnmFh/bj30Pk52PodF6kN3JYftvn
ZaXgOf6E4NLpHjtYRTzyVZQamenDAlvHQwZE284hDPShuJwxFr6FOSR/GrgqbN4d
cOK898ofM+ZxkNkm5LrU3RAXR3336HU9XMky4UCV9L3CA51IlTMqt/CkddFhsjrw
W4Zo1Aj8G7FaoDm7XhkLGDwVjf0Ua1O4YHRpSgVSkrXeBgW7P4Tc+53nFns3rwxs
uzF/x9tl2+BdiDjPOhSRuoa1ypilODdpOGKNKuf0vu2jAbbzDILBYOfw
-----END ENCRYPTED PRIVATE KEY-----"""

ORDEN_PAGO_ENDPOINT = '/ordenPago/registra'
CUENTA_ENDPOINT = '/cuentaModule/fisica'


def _desc_error(desc, id):
    return dict(resultado=dict(descripcionError=desc, id=id))


@pytest.mark.vcr
def test_forbidden_without_vpn(client):
    client = Client('TAMIZI', PKEY, '12345678', demo=False)
    with pytest.raises(HTTPError) as exc_info:
        client.request('get', '/application.wadl', {})
    assert exc_info.value.response.status_code == 403


def test_incorrect_passphrase():
    with pytest.raises(InvalidPassphrase):
        Client('TAMIZI', PKEY, 'incorrect')


@pytest.mark.parametrize(
    'client_mock,endpoint,expected_exc',
    [
        (
            _desc_error('No se recibió respuesta del servicio', 0),
            ORDEN_PAGO_ENDPOINT,
            NoServiceResponse,
        ),
        (
            _desc_error('Error validando la firma', 0),
            ORDEN_PAGO_ENDPOINT,
            SignatureValidationError,
        ),
        (
            _desc_error('El campo &lt;CONCEPTO PAGO> es obligatorio', 0),
            ORDEN_PAGO_ENDPOINT,
            MandatoryField,
        ),
        (
            _desc_error(
                'La clave de rastreo {foo123} para la fecha {20200314} de '
                'la institucion {123} ya fue utilizada',
                -1,
            ),
            ORDEN_PAGO_ENDPOINT,
            ClaveRastreoAlreadyInUse,
        ),
        (
            _desc_error('La cuenta {646180257067226640} no existe ', -7),
            ORDEN_PAGO_ENDPOINT,
            AccountDoesNotExist,
        ),
        (
            _desc_error('La Institucion 90679 no es valida', -9),
            ORDEN_PAGO_ENDPOINT,
            InvalidInstitution,
        ),
        (
            _desc_error('El tipo de cuenta 3 es invalido', -11),
            ORDEN_PAGO_ENDPOINT,
            InvalidAccountType,
        ),
        (
            _desc_error('El monto {500.0} no es válido', -20),
            ORDEN_PAGO_ENDPOINT,
            InvalidAmount,
        ),
        (
            _desc_error(
                'La cuenta CLABE {6461801570} no coincide para la '
                'institucion operante {40072}',
                -22,
            ),
            ORDEN_PAGO_ENDPOINT,
            BankCodeClabeMismatch,
        ),
        (
            _desc_error('Cuenta {646180157000000000} - {MISMA_CUENTA}', -24),
            ORDEN_PAGO_ENDPOINT,
            SameAccount,
        ),
        (
            _desc_error('Clave rastreo invalida : ABC123', -34),
            ORDEN_PAGO_ENDPOINT,
            InvalidTrackingKey,
        ),
        (
            _desc_error(
                'Orden sin cuenta ordenante. Se rechaza por PLD', -200
            ),
            ORDEN_PAGO_ENDPOINT,
            PldRejected,
        ),
        (
            _desc_error('unknown code', 9999999),
            ORDEN_PAGO_ENDPOINT,
            StpmexException,
        ),
        (
            dict(descripcion='Cuenta Duplicada', id=3),
            CUENTA_ENDPOINT,
            DuplicatedAccount,
        ),
        (
            dict(descripcion='El campo NOMBRE es invalido', id=1),
            CUENTA_ENDPOINT,
            InvalidField,
        ),
        (
            dict(descripcion='rfc/curp invalido', id=1),
            CUENTA_ENDPOINT,
            InvalidRfcOrCurp,
        ),
        (
            dict(descripcion='unknown code', id=999999),
            CUENTA_ENDPOINT,
            StpmexException,
        ),
        (
            dict(
                descripcion='El campo Apellido materno '
                'obligatorio 6461801500000000',
                id=5,
            ),
            CUENTA_ENDPOINT,
            MandatoryField,
        ),
    ],
    indirect=['client_mock'],
)
def test_errors(
    client_mock: Client, endpoint: str, expected_exc: type
) -> None:
    with pytest.raises(expected_exc) as exc_info:
        client_mock.put(endpoint, dict(firma='{hola}'))
    exc = exc_info.value
    assert repr(exc)
    assert str(exc)


@pytest.mark.vcr
def test_account_registration(client) -> None:
    client = Client('TAMIZI', PKEY, '12345678')
    response = client.put(CUENTA_ENDPOINT, dict(firma='{hola}'))
    assert response['id'] == 0
    assert response['descripcion'] == 'Cuenta en revisión.'



import datetime as dt
from dataclasses import asdict
from typing import Any, ClassVar, Dict, List

from ..auth import compute_signature, join_fields
from ..utils import strftime


class Resource:
    _client: ClassVar['stpmex.Client']  # noqa: F821
    _endpoint: ClassVar[str]
    _firma_fieldnames: ClassVar[List[str]]
    empresa: ClassVar[str]

    @property
    def firma(self):
        """
        Based on:
        https://stpmex.zendesk.com/hc/es/articles/360002796012-Firmas-Electr%C3%B3nicas-
        """
        joined_fields = join_fields(self, self._firma_fieldnames)
        return compute_signature(self._client.pkey, joined_fields)

    @classmethod
    def _firma_consulta(cls, consulta: Dict[str, Any]):
        joined = (
            f"|||"
            f"{cls.empresa}|"
            f"{consulta.get('fechaOperacion', '')}||"
            f"{consulta.get('claveRastreo', '')}|"
            f"{consulta.get('institucionOperante', '')}"
            f"||||||||||||||||||||||||||||||"
        )
        return compute_signature(cls._client.pkey, joined)

    def to_dict(self) -> Dict[str, Any]:
        base = dict()
        for k, v in asdict(self).items():
            if isinstance(v, dt.date):
                base[k] = strftime(v)
            elif v is not None:
                base[k] = v
        return {**base, **dict(firma=self.firma, empresa=self.empresa)}



import re
from typing import Any, ClassVar, Dict, List, NoReturn, Union

from cryptography.exceptions import UnsupportedAlgorithm
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from requests import Response, Session

from .exc import (
    AccountDoesNotExist,
    BankCodeClabeMismatch,
    ClaveRastreoAlreadyInUse,
    DuplicatedAccount,
    InvalidAccountType,
    InvalidAmount,
    InvalidField,
    InvalidInstitution,
    InvalidPassphrase,
    InvalidRfcOrCurp,
    InvalidTrackingKey,
    MandatoryField,
    NoOrdenesEncontradas,
    NoServiceResponse,
    PldRejected,
    SameAccount,
    SignatureValidationError,
    StpmexException,
)
from .resources import CuentaFisica, Orden, Resource, Saldo
from .version import __version__ as client_version

DEMO_HOST = 'https://demo.stpmex.com:7024'
PROD_HOST = 'https://prod.stpmex.com'


class Client:
    base_url: str
    soap_url: str
    session: Session

    # resources
    cuentas: ClassVar = CuentaFisica
    ordenes: ClassVar = Orden
    saldos: ClassVar = Saldo

    def __init__(
        self,
        empresa: str,
        priv_key: str,
        priv_key_passphrase: str,
        demo: bool = False,
        base_url: str = None,
        soap_url: str = None,
        timeout: tuple = None,
    ):
        self.timeout = timeout
        self.session = Session()
        self.session.headers['User-Agent'] = f'stpmex-python/{client_version}'
        if demo:
            host_url = DEMO_HOST
            self.session.verify = False
        else:
            host_url = PROD_HOST
            self.session.verify = True
        self.base_url = base_url or f'{host_url}/speiws/rest'
        self.soap_url = (
            soap_url or f'{host_url}/spei/webservices/SpeiConsultaServices'
        )

        try:
            self.pkey = serialization.load_pem_private_key(
                priv_key.encode('utf-8'),
                priv_key_passphrase.encode('ascii'),
                default_backend(),
            )
        except (ValueError, TypeError, UnsupportedAlgorithm):
            raise InvalidPassphrase
        Resource.empresa = empresa
        Resource._client = self

    def post(
        self, endpoint: str, data: Dict[str, Any]
    ) -> Union[Dict[str, Any], List[Any]]:
        return self.request('post', endpoint, data)

    def put(
        self, endpoint: str, data: Dict[str, Any]
    ) -> Union[Dict[str, Any], List[Any]]:
        return self.request('put', endpoint, data)

    def delete(
        self, endpoint: str, data: Dict[str, Any]
    ) -> Union[Dict[str, Any], List[Any]]:
        return self.request('delete', endpoint, data)

    def request(
        self, method: str, endpoint: str, data: Dict[str, Any], **kwargs: Any
    ) -> Union[Dict[str, Any], List[Any]]:
        url = self.base_url + endpoint
        response = self.session.request(
            method,
            url,
            json=data,
            timeout=self.timeout,
            **kwargs,
        )
        self._check_response(response)
        resultado = response.json()
        if 'resultado' in resultado:  # Some responses are enveloped
            resultado = resultado['resultado']
        return resultado

    @staticmethod
    def _check_response(response: Response) -> None:
        if not response.ok:
            response.raise_for_status()
        resp = response.json()
        if isinstance(resp, dict):
            try:
                _raise_description_error_exc(resp)
            except KeyError:
                ...
            try:
                assert resp['descripcion']
                _raise_description_exc(resp)
            except (AssertionError, KeyError):
                ...
        response.raise_for_status()


def _raise_description_error_exc(resp: Dict) -> NoReturn:
    id = resp['resultado']['id']
    error = resp['resultado']['descripcionError']

    if id == 0 and error == 'No se recibió respuesta del servicio':
        raise NoServiceResponse(**resp['resultado'])
    elif id == 0 and error == 'Error validando la firma':
        raise SignatureValidationError(**resp['resultado'])
    elif id == 0 and re.match(r'El campo .+ es obligatorio', error):
        raise MandatoryField(**resp['resultado'])
    elif id == -1 and re.match(
        r'La clave de rastreo .+ ya fue utilizada', error
    ):
        raise ClaveRastreoAlreadyInUse(**resp['resultado'])
    elif id == -7 and re.match(r'La cuenta .+ no existe', error):
        raise AccountDoesNotExist(**resp['resultado'])
    elif id == -9 and re.match(r'La Institucion \d+ no es valida', error):
        raise InvalidInstitution(**resp['resultado'])
    elif id == -11 and re.match(r'El tipo de cuenta \d+ es invalido', error):
        raise InvalidAccountType(**resp['resultado'])
    elif id == -20 and re.match(r'El monto {.+} no es válido', error):
        raise InvalidAmount(**resp['resultado'])
    elif id == -22 and 'no coincide para la institucion operante' in error:
        raise BankCodeClabeMismatch(**resp['resultado'])
    elif id == -24 and re.match(r'Cuenta {\d+} - {MISMA_CUENTA}', error):
        raise SameAccount(**resp['resultado'])
    elif id == -34 and 'Clave rastreo invalida' in error:
        raise InvalidTrackingKey(**resp['resultado'])
    elif id == -100 and error.startswith('No se encontr'):
        raise NoOrdenesEncontradas
    elif id == -200 and 'Se rechaza por PLD' in error:
        raise PldRejected(**resp['resultado'])
    else:
        raise StpmexException(**resp['resultado'])


def _raise_description_exc(resp: Dict) -> NoReturn:
    id = resp['id']
    desc = resp['descripcion']

    if id == 0 and 'Cuenta en revisión' in desc:
        # STP regresa esta respuesta cuando se registra
        # una cuenta. No se levanta excepción porque
        # todas las cuentas pasan por este status.
        ...
    elif id == 1 and desc == 'rfc/curp invalido':
        raise InvalidRfcOrCurp(**resp)
    elif id == 1 and re.match(r'El campo \w+ es invalido', desc):
        raise InvalidField(**resp)
    elif id == 3 and desc == 'Cuenta Duplicada':
        raise DuplicatedAccount(**resp)
    elif id == 5 and re.match(r'El campo .* obligatorio \w+', desc):
        raise MandatoryField(**resp)
    else:
        raise StpmexException(**resp)



from importlib.machinery import SourceFileLoader

from setuptools import find_packages, setup

version = SourceFileLoader('version', 'stpmex/version.py').load_module()

install_requires = [
    'cryptography>=3.0,<3.5',
    'cuenca-validations>=0.4,<0.10',
    'requests>=2.25',
]


with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='stpmex',
    version=version.__version__,
    author='Cuenca',
    author_email='dev@cuenca.com',
    description='Client library for stpmex.com',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/cuenca-mx/stpmex-python',
    packages=find_packages(),
    include_package_data=True,
    package_data=dict(stpmex=['py.typed']),
    python_requires='>=3.6',
    install_requires=install_requires,
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)
